use std::collections::HashSet;

use axum::{extract::State, response::IntoResponse, Json};
use tokio::select;

use crate::{
    error::{AppError, AppResult},
    models::{AppStateShared, UrlReportResponse, UrlRequest, UrlResult},
    repo::Repo,
};

pub async fn get_urls<R: Repo>(
    State(state): State<AppStateShared<R>>,
    req: Option<Json<UrlRequest>>,
) -> AppResult<impl IntoResponse> {
    let urls = req.map(|urls| {
        urls.0
            .urls
            .into_iter()
            .map(|u| u.to_string())
            .collect::<HashSet<_>>()
    });
    let r_state = state.read().await;
    let result = r_state.repo.get_urls(urls.as_ref())?;

    Ok(Json(result))
}

pub async fn upload_urls<R: Repo>(
    State(state): State<AppStateShared<R>>,
    Json(payload): Json<UrlRequest>,
) -> AppResult<impl IntoResponse> {
    let mut successes = vec![];
    let mut failures = vec![];

    let urls = {
        let r_state = state.read().await;
        let cached_urls = r_state.repo.get_urls(None)?;

        payload
            .urls
            .into_iter()
            .filter(|url| match cached_urls.get(&url.to_string()) {
                Some(val) => {
                    info!("{} is cached, not fetching", url);
                    successes.push(UrlResult {
                        status: None,
                        url: url.clone(),
                        content_length: *val,
                        error_msg: None,
                    });
                    false
                }
                None => true,
            })
            .collect::<HashSet<_>>()
    };

    let tasks = {
        urls.into_iter()
            .map(|url| {
                let state = state.clone();
                tokio::spawn(async move {
                    let r_state = state.read().await;
                    let _sem = r_state.url_handling_semaphore.acquire().await.unwrap();
                    let ct = r_state.cancellation_token.clone();
                    info!("fetching {url}");

                    let request = (|| async {
                        let resp = reqwest::get(url.to_string()).await?;
                        let status = resp.status().as_u16();
                        let result = resp.text().await?;

                        Ok((status, result))
                    })();

                    let response: AppResult<_> = select! {
                        r = request => r,
                        _ = ct.cancelled() => {
                            info!("fetching {url} cancelled");
                            Err(AppError::Cancelled(url.to_string()))
                        }
                    };

                    match response {
                        Ok((status, resp)) => UrlResult {
                            status: Some(status),
                            url,
                            error_msg: None,
                            content_length: resp.len(),
                        },
                        Err(err) => UrlResult {
                            status: None,
                            url,
                            error_msg: Some(err.to_string()),
                            content_length: 0,
                        },
                    }
                })
            })
            .collect::<Vec<_>>()
    };

    let ct = state.read().await.cancellation_token.clone();
    for task in tasks {
        let result = if let Some(result) = ct.run_until_cancelled(task).await {
            result.unwrap()
        } else {
            continue;
        };

        let status = result.status.unwrap_or(0);

        if status >= 200 && status < 400 {
            successes.push(result);
        } else {
            failures.push(result);
        }
    }
    info!("fetching completed");

    let mut w_state = state.write().await;
    w_state.repo.insert_urls(&successes)?;
    w_state.regen_cancellation_token();
    drop(w_state);

    Ok(Json(UrlReportResponse {
        successes,
        failures,
    }))
}

pub async fn delete_urls<R: Repo>(
    State(state): State<AppStateShared<R>>,
    req: Option<Json<UrlRequest>>,
) -> AppResult<impl IntoResponse> {
    let urls = req.map(|urls| {
        urls.0
            .urls
            .into_iter()
            .map(|u| u.to_string())
            .collect::<HashSet<_>>()
    });

    let deleted = state.write().await.repo.delete_urls(urls.as_ref())?;

    Ok(deleted.to_string())
}
