mod configs;
mod error;
mod handlers;
mod models;
mod repo;

use std::sync::Arc;

use axum::{routing::post, Router};

use error::{AppError, AppResult};
use models::AppState;
use repo::Repo;
use tokio::{net::TcpListener, sync::RwLock};
use tokio_util::sync::CancellationToken;

#[macro_use]
extern crate log;

async fn run() -> AppResult<()> {
    dotenv::dotenv().ok();

    let cfg = configs::get_configs()?;
    info!("{cfg:?}");

    let repo = create_repo()?;
    let ct = CancellationToken::new();
    let state = Arc::new(RwLock::new(AppState::new(
        cfg.url_handlers_pool_size,
        repo,
        ct.clone(),
    )));

    #[cfg(unix)]
    {
        use tokio::signal::unix::{signal, SignalKind};

        let state = state.clone();
        tokio::spawn(async move {
            let mut signal = signal(SignalKind::hangup()).unwrap();
            loop {
                signal.recv().await;
                state.read().await.cancellation_token.cancel();
                info!("SIGHUP");
            }
        });
    }

    let app = Router::new()
        .route(
            "/urls",
            post(|state, body| handlers::get_urls(state, Some(body)))
                .get(|state| handlers::get_urls(state, None)),
        )
        .route("/urls/upload", post(handlers::upload_urls))
        .route(
            "/urls/delete",
            post(|state, req| handlers::delete_urls(state, Some(req)))
                .get(|state| handlers::delete_urls(state, None)),
        )
        .with_state(state);

    let bind_addr = format!("0.0.0.0:{}", cfg.port);
    let listener = TcpListener::bind(bind_addr).await?;

    axum::serve(listener, app).await.map_err(AppError::from)
}

fn create_repo() -> AppResult<impl Repo> {
    let config = sled::Config::new().temporary(true);
    config.open().map_err(AppError::from)
}

#[tokio::main]
async fn main() {
    env_logger::init();

    let _ = run().await.inspect_err(|e| println!("{e}"));
}
